#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>

#include <fftw3.h>

#ifdef _OPENMP
#include <omp.h>
#endif

// HACC
#include "complex-type.h"

#define FFTW_ADDR(X) reinterpret_cast<fftw_complex*>(&(X)[0])

double mygettimeofday() {
  timeval tv;
  gettimeofday(&tv, NULL);
  return 1.0*tv.tv_sec + 1.0e-6 * tv.tv_usec;
}

class FFT1D {
public:

  FFT1D(int size, int howmany, unsigned flags, bool OutOfPlace) :
    size(size),
    howmany(howmany),
    flags(flags),
    OutOfPlace(OutOfPlace)
  {
    n_elem = (size_t)size * (size_t)howmany;
    ra = (complex_t *)fftw_malloc(n_elem*sizeof(complex_t));
    if(OutOfPlace)
      ka = (complex_t *)fftw_malloc(n_elem*sizeof(complex_t));
    else
      ka = ra;
    double t0 = mygettimeofday();
    makePlans();
    double t1 = mygettimeofday();
    t_plan = t1 - t0;
  }

  ~FFT1D() {
    fftw_free(ra);
    if(OutOfPlace)
      fftw_free(ka);
    fftw_destroy_plan(plan_f);
    fftw_destroy_plan(plan_b);
  }
  
  void makePlans() {
    const int n[1] = {size};

    // forward FFT from ra to ka
    plan_f = fftw_plan_many_dft(1,             // int rank,
				n,             // const int *n,
				howmany,       // int howmany,
				FFTW_ADDR(ra), // fftw_complex *in,
				NULL,          // const int *inembed,
				1,             // int istride,
				size,          // int idist,
				FFTW_ADDR(ka), // fftw_complex *out,
				NULL,          // const int *onembed,
				1,             // int ostride,
				size,          // int odist,
				FFTW_FORWARD,  // int sign,
				flags);        // unsigned flags

    // backward FFT from ka to ra
    plan_b = fftw_plan_many_dft(1,             // int rank,
				&n[0],         // const int *n,
				howmany,       // int howmany,
				FFTW_ADDR(ka), // fftw_complex *in,
				NULL,          // const int *inembed,
				1,             // int istride,
				size,          // int idist,
				FFTW_ADDR(ra), // fftw_complex *out,
				NULL,          // const int *onembed,
				1,             // int ostride,
				size,          // int odist,
				FFTW_BACKWARD, // int sign,
				flags);        // unsigned flags
  }

  void setupDirac() {
    complex_t zero(0.0, 0.0);
    complex_t one(1.0, 0.0);
    
    // fill r-space array with zeros
#pragma omp parallel for
    for(size_t i=0; i<n_elem; i++) {
      ra[i] = zero;
    }
    
    // put one at beginning of each r-space sub-array
#pragma omp parallel for
    for(size_t j=0; j<howmany; j++) {
      ra[j*size] = one;
    }

    // fill k space array with zeros
    if(OutOfPlace) {
#pragma omp parallel for
      for(size_t i=0; i<n_elem; i++) {
	ka[i] = zero;
      }
    }
  }

  size_t checkDiracR() {
    complex_t zero(0.0, 0.0);
    complex_t one(1.0, 0.0);

    // pick up a factor of ng after forward/backward
    one *= (1.0*size);
    
    size_t nerrors=0;
#pragma omp parallel for reduction(+:nerrors)
    for(size_t j=0; j<howmany; j++) {
      size_t indx = j*size;
      nerrors += (ra[indx] != one);
      indx++;
      for(size_t i=1; i<size; i++) {
	nerrors += (ra[indx] != zero);
	indx++;
      }
    }

    return nerrors;
  }

  size_t checkDiracK() {
    complex_t one(1.0, 0.0);

    size_t nerrors=0;
#pragma omp parallel for reduction(+:nerrors)
    for(size_t i=0; i<n_elem; i++)
      nerrors += (ka[i] != one);

    return nerrors;
  }

  size_t testDirac(int repetitions=1) {
    for(size_t rep=0; rep<repetitions; rep++) {

      printf("%zu\n",rep);

      double t0, t1;
      size_t nerrors;

      // setup arrays for Dirac delta test
      t0 = mygettimeofday();
      setupDirac();
      t1 = mygettimeofday();
      printf("t_setup = %f seconds\n", t1-t0);

      // forward FFT
      t0 = mygettimeofday();
      fftw_execute(plan_f);
      t1 = mygettimeofday();
      printf("t_fftf = %f seconds\n", t1-t0);

      // check array in k-space
      t0 = mygettimeofday();
      nerrors = checkDiracK();
      t1 = mygettimeofday();
      printf("t_check = %f seconds\n", t1-t0);
      printf("n_errors = %zu\n",nerrors);

      // backward FFT
      t0 = mygettimeofday();
      fftw_execute(plan_b);
      t1 = mygettimeofday();
      printf("t_fftb = %f seconds\n", t1-t0);

      // check array in r-space
      t0 = mygettimeofday();
      nerrors = checkDiracR();
      t1 = mygettimeofday();
      printf("t_check = %f seconds\n", t1-t0);
      printf("n_errors = %zu\n",nerrors);

      printf("\n");
    }

    return 0;
  }
  
  int size, howmany;
  unsigned flags;
  bool OutOfPlace;
  size_t n_elem;
  complex_t *ra, *ka;
  fftw_plan plan_f, plan_b;
  double t_plan;
};

int main(int argc, char *argv[]) {

  if(argc < 6) {
    printf("USAGE: %s <size> <howmany> <repetitions> <FFTW_MEASURE=0or1> <OutOfPlace=0or1>\n", argv[0]);
    exit(-1);
  }
  int size = atoi(argv[1]);
  int howmany = atoi(argv[2]);
  int repetitions = atoi(argv[3]);
  int measureq = atoi(argv[4]);
  int oopq = atoi(argv[5]);

  unsigned flags = FFTW_ESTIMATE;
  if(measureq) {
    flags = FFTW_MEASURE;
    printf("FFTW_MEASURE\n");
  } else {
    printf("FFTW_ESTIMATE\n");
  }
  
  // set up omp threading
#ifdef _OPENMP
  int nthreads = omp_get_max_threads();
  printf("\nomp_max_threads = %d\n",nthreads);
  fftw_init_threads();
  fftw_plan_with_nthreads(nthreads);
  printf("fftw_planner_nthreads = %d\n\n",fftw_planner_nthreads());
#endif

  // consider separate executions for ESTIMATE vs MEASURE?
  // ESTIMATE before MEASURE is reasonably fair
  
  // FFTW_ESTIMATE first otherwise FFTW3 remembers from FFTW_MEASURE
  FFT1D *fft1d = new FFT1D(size, howmany, flags, oopq != 0);
  printf("t_plan = %f seconds\n\n", fft1d->t_plan);
  fft1d->testDirac(repetitions);
  delete fft1d;

#ifdef _OPENMP
  fftw_cleanup_threads();
#else
  fftw_cleanup();
#endif
  
  return 0;
}
